# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import socket

import kombu
from celery import current_app
from django.conf import settings as django_settings
from redis.exceptions import ConnectionError

from pipeline.celery.settings import CELERY_QUEUES
from pipeline.conf import settings
from pipeline.django_signal_valve import valve
from pipeline.engine import signals
from pipeline.engine.core import data
from pipeline.engine.exceptions import RabbitMQConnectionError
from pipeline.engine.models import FunctionSwitch, PipelineProcess

logger = logging.getLogger("root")
WORKER_PING_TIMES = 2


def freeze():
    # turn on switch
    FunctionSwitch.objects.freeze_engine()


def unfreeze():
    # turn off switch
    FunctionSwitch.objects.unfreeze_engine()

    # resend signal
    valve.open_valve(signals)

    # unfreeze process
    frozen_process_list = PipelineProcess.objects.filter(is_frozen=True)
    for process in frozen_process_list:
        process.unfreeze()


def workers(connection=None):
    try:
        worker_list = data.cache_for("__pipeline__workers__")
    except ConnectionError as e:
        logger.exception("pipeline cache_for __pipeline__workers__ raise error: %s" % e)
        raise e

    if not worker_list:
        tries = 0
        while tries < WORKER_PING_TIMES:
            kwargs = {"timeout": tries + 1}
            if connection is not None:
                kwargs["connection"] = connection
            try:
                worker_list = current_app.control.ping(**kwargs)
            except socket.error as err:
                logger.exception("pipeline current_app.control.ping raise error: %s" % err)
                # raise error at last loop
                if tries >= WORKER_PING_TIMES - 1:
                    raise RabbitMQConnectionError(err)

            if worker_list:
                break

            tries += 1

        if worker_list:
            data.expire_cache("__pipeline__workers__", worker_list, settings.PIPELINE_WORKER_STATUS_CACHE_EXPIRES)

    return worker_list


def stats():
    inspect = current_app.control.inspect()

    stats = {"workers": {}, "queues": {}}

    worker_stats = inspect.stats()
    active_queues = inspect.active_queues()

    if worker_stats:

        for name, stat in worker_stats.items():
            stats["workers"].setdefault(name, {"stat": {}, "queues": {}})["stat"] = stat

    if active_queues:

        for name, queues in active_queues.items():
            stats["workers"].setdefault(name, {"stat": {}, "queues": {}})["queues"] = queues

    if not hasattr(django_settings, "BROKER_VHOST"):
        stats["queues"] = "can not find BROKER_VHOST in django settings"

        return stats

    with kombu.Connection(django_settings.BROKER_URL) as conn:
        client = conn.get_manager()

        if not hasattr(client, "get_queue"):
            stats["queues"] = "broker does not support queues info query"

            return stats

        for queue in CELERY_QUEUES:
            stats["queues"][queue.name] = client.get_queue(django_settings.BROKER_VHOST, queue.name)

    return stats



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import warnings
from functools import wraps
from itertools import count

from django.conf import settings
from django.db import connection, models
from django.db.models.query import QuerySet

from pipeline.contrib.periodic_task.djcelery.db import (
    commit_on_success,
    get_queryset,
    rollback_unless_managed,
)
from pipeline.contrib.periodic_task.djcelery.utils import now

try:
    from django.db import connections, router
except ImportError:  # pre-Django 1.2
    connections = router = None  # noqa


try:
    from celery.utils.timeutils import maybe_timedelta
except ImportError:
    from celery.utils.time import maybe_timedelta


def update_model_with_dict(obj, fields):
    [setattr(obj, attr_name, attr_value) for attr_name, attr_value in list(fields.items())]
    obj.save()
    return obj


class TxIsolationWarning(UserWarning):
    pass


def transaction_retry(max_retries=1):
    """Decorator for methods doing database operations.

    If the database operation fails, it will retry the operation
    at most ``max_retries`` times.

    """

    def _outer(fun):
        @wraps(fun)
        def _inner(*args, **kwargs):
            _max_retries = kwargs.pop("exception_retry_count", max_retries)
            for retries in count(0):
                try:
                    return fun(*args, **kwargs)
                except Exception:  # pragma: no cover
                    # Depending on the database backend used we can experience
                    # various exceptions. E.g. psycopg2 raises an exception
                    # if some operation breaks the transaction, so saving
                    # the task result won't be possible until we rollback
                    # the transaction.
                    if retries >= _max_retries:
                        raise
                    try:
                        rollback_unless_managed()
                    except Exception:
                        pass

        return _inner

    return _outer


class ExtendedQuerySet(QuerySet):
    def update_or_create(self, **kwargs):
        obj, created = self.get_or_create(**kwargs)

        if not created:
            fields = dict(kwargs.pop("defaults", {}))
            fields.update(kwargs)
            update_model_with_dict(obj, fields)

        return obj, created


class ExtendedManager(models.Manager):
    def get_queryset(self):
        return ExtendedQuerySet(self.model)

    get_query_set = get_queryset  # Pre django 1.6

    def update_or_create(self, **kwargs):
        return get_queryset(self).update_or_create(**kwargs)

    def connection_for_write(self):
        if connections:
            return connections[router.db_for_write(self.model)]
        return connection

    def connection_for_read(self):
        if connections:
            return connections[self.db]
        return connection

    def current_engine(self):
        try:
            return settings.DATABASES[self.db]["ENGINE"]
        except AttributeError:
            return settings.DATABASE_ENGINE


class ResultManager(ExtendedManager):
    def get_all_expired(self, expires):
        """Get all expired task results."""
        return self.filter(date_done__lt=now() - maybe_timedelta(expires))

    def delete_expired(self, expires):
        """Delete all expired taskset results."""
        meta = self.model._meta
        with commit_on_success():
            self.get_all_expired(expires).update(hidden=True)
            cursor = self.connection_for_write().cursor()
            cursor.execute(
                "DELETE FROM {0.db_table} WHERE hidden=%s".format(meta), (True,),
            )


class PeriodicTaskManager(ExtendedManager):
    def enabled(self):
        return self.filter(enabled=True)


class TaskManager(ResultManager):
    """Manager for :class:`celery.models.Task` models."""

    _last_id = None

    def get_task(self, task_id):
        """Get task meta for task by ``task_id``.

        :keyword exception_retry_count: How many times to retry by
            transaction rollback on exception. This could theoretically
            happen in a race condition if another worker is trying to
            create the same task. The default is to retry once.

        """
        try:
            return self.get(task_id=task_id)
        except self.model.DoesNotExist:
            if self._last_id == task_id:
                self.warn_if_repeatable_read()
            self._last_id = task_id
            return self.model(task_id=task_id)

    @transaction_retry(max_retries=2)
    def store_result(self, task_id, result, status, traceback=None, children=None):
        """Store the result and status of a task.

        :param task_id: task id

        :param result: The return value of the task, or an exception
            instance raised by the task.

        :param status: Task status. See
            :meth:`celery.result.AsyncResult.get_status` for a list of
            possible status values.

        :keyword traceback: The traceback at the point of exception (if the
            task failed).

        :keyword children: List of serialized results of subtasks
            of this task.

        :keyword exception_retry_count: How many times to retry by
            transaction rollback on exception. This could theoretically
            happen in a race condition if another worker is trying to
            create the same task. The default is to retry twice.

        """
        return self.update_or_create(
            task_id=task_id,
            defaults={"status": status, "result": result, "traceback": traceback, "meta": {"children": children}},
        )

    def warn_if_repeatable_read(self):
        if "mysql" in self.current_engine().lower():
            cursor = self.connection_for_read().cursor()
            if cursor.execute("SELECT @@tx_isolation"):
                isolation = cursor.fetchone()[0]
                if isolation == "REPEATABLE-READ":
                    warnings.warn(
                        TxIsolationWarning(
                            "Polling results with transaction isolation level "
                            "repeatable-read within the same transaction "
                            "may give outdated results. Be sure to commit the "
                            "transaction for each poll iteration."
                        )
                    )



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import time

from pipeline.core.data.base import DataObject
from pipeline.utils.uniqid import uniqid


def get_console_logger():
    # create logger
    logger = logging.getLogger("simple_example")
    logger.setLevel(logging.DEBUG)

    # create console handler and set level to debug
    ch = logging.StreamHandler()
    ch.setLevel(logging.DEBUG)

    # create formatter
    formatter = logging.Formatter("%(asctime)s - %(levelname)s - %(message)s")

    # add formatter to ch
    ch.setFormatter(formatter)

    # add ch to logger
    logger.addHandler(ch)

    return logger


logger = get_console_logger()


class ComponentRunner:
    def __init__(self, component_cls):
        self.component_cls = component_cls

    def run(self, data, parent_data, callback_data=None):

        service = self.component_cls.bound_service()

        setattr(service, "id", uniqid())
        setattr(service, "logger", logger)

        data_object = DataObject(inputs=data)
        parent_data_object = DataObject(inputs=parent_data)

        logger.info(
            "Start to run component [{}] with data: {}, parent_data: {}".format(
                self.component_cls.code, data_object, parent_data_object
            )
        )

        result = service.execute(data_object, parent_data_object)

        if result is False:
            logger.info("Execute return [{}], stop running.".format(result))
            return

        if not service.need_schedule():
            logger.info("Execute return [{}], and component do not need schedule, finish running".format(result))
            return

        if service.interval is None:
            logger.info("Start to callback component with callbackdata: {}".format(callback_data))
            result = service.schedule(data_object, parent_data_object, callback_data)

            if result is False:
                logger.info("Schedule return [{}], stop running.".format(result))
                return
            else:
                logger.info("Schedule return [{}], finish running".format(result))
        else:

            schedue_times = 0

            while not service.is_schedule_finished():

                schedue_times += 1

                logger.info(
                    "Schedule {} with data: {}, parent_data: {}".format(schedue_times, data_object, parent_data_object)
                )

                result = service.schedule(data_object, parent_data_object, None)

                if result is False:
                    logger.info("Schedule return [{}], stop running.".format(result))
                    return

                interval = service.interval.next()
                logger.info("Schedule return [{}], wait for next schedule in {}s".format(result, interval))
                time.sleep(interval)

            logger.info("Schedule finished")



# Generated by Django 2.2.28 on 2022-04-22 02:48

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('eri', '0004_state_inner_loop_'),
    ]

    operations = [
        migrations.AlterField(
            model_name='contextvalue',
            name='code',
            field=models.CharField(blank=True, default='', max_length=128, verbose_name='计算型变量类型唯一标志'),
        ),
        migrations.AlterField(
            model_name='logentry',
            name='message',
            field=models.TextField(blank=True, null=True, verbose_name='日志内容'),
        ),
        migrations.AlterField(
            model_name='logentry',
            name='version',
            field=models.CharField(blank=True, default='', max_length=33, verbose_name='状态版本'),
        ),
        migrations.AlterField(
            model_name='process',
            name='destination_id',
            field=models.CharField(blank=True, default='', max_length=33, verbose_name='执行终点 ID'),
        ),
        migrations.AlterField(
            model_name='process',
            name='queue',
            field=models.CharField(blank=True, default='', max_length=128, verbose_name='所属队列'),
        ),
        migrations.AlterField(
            model_name='process',
            name='suspended_by',
            field=models.CharField(blank=True, db_index=True, default='', max_length=33, verbose_name='导致进程暂停的节点 ID'),
        ),
        migrations.AlterField(
            model_name='state',
            name='archived_time',
            field=models.DateTimeField(blank=True, null=True, verbose_name='归档时间'),
        ),
        migrations.AlterField(
            model_name='state',
            name='parent_id',
            field=models.CharField(blank=True, db_index=True, default='', max_length=33, verbose_name='父节点 ID'),
        ),
        migrations.AlterField(
            model_name='state',
            name='root_id',
            field=models.CharField(blank=True, db_index=True, default='', max_length=33, verbose_name='根节点 ID'),
        ),
        migrations.AlterField(
            model_name='state',
            name='started_time',
            field=models.DateTimeField(blank=True, null=True, verbose_name='开始时间'),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import _ast
import ast

from django.utils.module_loading import import_string

from .exceptions import ForbiddenMakoTemplateException


class StrictNodeVisitor(ast.NodeVisitor):
    """
    遍历语法树节点，遇到魔术方法使用或import时，抛出异常
    """

    BLACK_LIST_MODULE_METHODS = {
        "os": dir(__import__("os")),
        "subprocess": dir(__import__("subprocess")),
        "shutil": dir(__import__("shutil")),
        "ctypes": dir(__import__("ctypes")),
        "codecs": dir(__import__("codecs")),
        "sys": dir(__import__("sys")),
        "socket": dir(__import__("socket")),
        "webbrowser": dir(__import__("webbrowser")),
        "threading": dir(__import__("threading")),
        "sqlite3": dir(__import__("threading")),
        "signal": dir(__import__("signal")),
        "imaplib": dir(__import__("imaplib")),
        "fcntl": dir(__import__("fcntl")),
        "pdb": dir(__import__("pdb")),
        "pty": dir(__import__("pty")),
        "glob": dir(__import__("glob")),
        "tempfile": dir(__import__("tempfile")),
        "types": dir(import_string("types.CodeType")) + dir(import_string("types.FrameType")),
        "builtins": [
            "getattr",
            "hasattr",
            "breakpoint",
            "compile",
            "delattr",
            "open",
            "eval",
            "exec",
            "execfile",
            "exit",
            "dir",
            "globals",
            "locals",
            "input",
            "iter",
            "next",
            "quit",
            "setattr",
            "vars",
            "memoryview",
            "super",
            "print",
        ],
    }

    BLACK_LIST_METHODS = []
    for module_name, methods in BLACK_LIST_MODULE_METHODS.items():
        BLACK_LIST_METHODS.append(module_name)
        BLACK_LIST_METHODS.extend(methods)
    BLACK_LIST_METHODS = set(BLACK_LIST_METHODS)

    WHITE_LIST_MODULES = ["datetime", "re", "random", "json", "math"]

    def __init__(self, black_list_methods=None, white_list_modules=None):
        self.black_list_methods = black_list_methods or self.BLACK_LIST_METHODS
        self.white_list_modules = white_list_modules or self.WHITE_LIST_MODULES

    @staticmethod
    def is_white_list_ast_obj(ast_obj: _ast.AST) -> bool:
        """
        判断是否白名单对象，特殊豁免
        :param ast_obj: 抽象语法树节点
        :return: bool
        """
        # re 正则表达式允许使用 compile
        if isinstance(ast_obj, _ast.Attribute) and isinstance(ast_obj.value, _ast.Name):
            if ast_obj.value.id == "re" and ast_obj.attr in ["compile"]:
                return True

        return False

    def visit_Attribute(self, node):
        if self.is_white_list_ast_obj(node):
            return

        if node.attr in self.black_list_methods or node.attr.startswith("_"):
            raise ForbiddenMakoTemplateException("Mako template forbidden.")

    def visit_Name(self, node):
        if node.id in self.black_list_methods or node.id.startswith("_"):
            raise ForbiddenMakoTemplateException("Mako template forbidden.")

    def visit_Import(self, node):
        for name in node.names:
            if name.name not in self.white_list_modules:
                raise ForbiddenMakoTemplateException("Mako template forbidden.")

    def visit_ImportFrom(self, node):
        self.visit_Import(node)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback
from abc import ABCMeta, abstractmethod

from pipeline.core.flow.base import FlowNode
from pipeline.core.pipeline import Pipeline
from pipeline.engine.signals import pipeline_end

logger = logging.getLogger("celery")


class Event(FlowNode, metaclass=ABCMeta):
    def __init__(self, id, name=None, data=None):
        super(Event, self).__init__(id, name, data)

    def next(self):
        return self.outgoing.unique_one().target


class ThrowEvent(Event, metaclass=ABCMeta):
    pass


class CatchEvent(Event, metaclass=ABCMeta):
    pass


class EndEvent(ThrowEvent, metaclass=ABCMeta):
    def pipeline_finish(self, root_pipeline_id):
        try:
            pipeline_end.send(sender=Pipeline, root_pipeline_id=root_pipeline_id)
        except Exception:
            logger.error("pipeline end handler error %s" % traceback.format_exc())


class StartEvent(CatchEvent, metaclass=ABCMeta):
    pass


class EmptyStartEvent(StartEvent):
    pass


class EmptyEndEvent(EndEvent):
    pass


class ExecutableEndEvent(EndEvent, metaclass=ABCMeta):
    @abstractmethod
    def execute(self, in_subprocess, root_pipeline_id, current_pipeline_id):
        raise NotImplementedError()



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import weakref
from abc import ABCMeta, abstractmethod
from functools import wraps

from pipeline.exceptions import InvalidOperationException


def extra_inject(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        if "extra" not in kwargs:
            kwargs["extra"] = {}
        kwargs["extra"]["_id"] = args[0].id
        return func(*args, **kwargs)

    return wrapper


class FlowElement(object, metaclass=ABCMeta):
    def __init__(self, id, name=None):
        self.id = id
        self.name = name


class FlowNode(FlowElement, metaclass=ABCMeta):
    ON_RETRY = "_on_retry"

    def __init__(self, id, name=None, data=None):
        super(FlowNode, self).__init__(id, name)
        self.incoming = SequenceFlowCollection()
        self.outgoing = SequenceFlowCollection()
        self.data = data

    def on_retry(self):
        return hasattr(self, self.ON_RETRY)

    def next_exec_is_retry(self):
        setattr(self, self.ON_RETRY, True)

    def retry_at_current_exec(self):
        delattr(self, self.ON_RETRY)

    @abstractmethod
    def next(self):
        """
        该节点的下一个节点，由子类来实现
        :return:
        """
        raise NotImplementedError()

    class FlowNodeLogger:
        def __init__(self, id):
            self.id = id
            self._logger = logging.getLogger("pipeline.logging")

        @extra_inject
        def info(self, *args, **kwargs):
            self._logger.info(*args, **kwargs)

        @extra_inject
        def warning(self, *args, **kwargs):
            self._logger.warning(*args, **kwargs)

        @extra_inject
        def error(self, *args, **kwargs):
            self._logger.error(*args, **kwargs)

        @extra_inject
        def critical(self, *args, **kwargs):
            self._logger.critical(*args, **kwargs)

    @property
    def logger(self):
        _logger = getattr(self, "_logger", None)
        if not _logger:
            _logger = self.FlowNodeLogger(self.id)
            setattr(self, "_logger", _logger)
        return _logger

    def __getstate__(self):
        if "_logger" in self.__dict__:
            del self.__dict__["_logger"]
        return self.__dict__


class SequenceFlow(FlowElement):
    def __init__(self, id, source, target, is_default=False, name=None):
        super(SequenceFlow, self).__init__(id, name)
        self.source = weakref.proxy(source) if source is not None else source
        self.target = weakref.proxy(target) if target is not None else target
        self.is_default = is_default


class SequenceFlowCollection(object):
    def __init__(self, *flows):
        flow_dict = {}
        for flow in flows:
            flow_dict[flow.id] = flow

        self.flows = list(flows)
        self.flow_dict = flow_dict

    def get_flow(self, id):
        """
        获取 flow.id = id 的某个 flow
        :param id: flow id
        :return:
        """
        return self.flow_dict.get(id)

    def unique_one(self):
        """
        获取唯一的一个 flow，若当前集合内 flow 不只一条则抛出异常
        :return:
        """
        if len(self.flows) != 1:
            raise InvalidOperationException("this collection contains multiple flow, can not get unique one.")
        return self.flows[0]

    def is_empty(self):
        """
        当前集合是否为空
        :return:
        """
        return len(self.flows) == 0

    def default_flow(self):
        """
        获取当前集合中默认的 flow
        :return: 若存在默认的 flow 则返回，否则返回 None
        """
        for flow in self.flows:
            if flow.is_default:
                return flow
        return None

    def add_flow(self, flow):
        """
        向当前结合中添加一条 flow
        :param flow: 待添加的 flow
        :return:
        """
        self.flows.append(flow)
        self.flow_dict[flow.id] = flow

    def all_target_node(self):
        """
        返回当前集合中所有 flow 的 target
        :return:
        """
        nodes = []
        for flow in self.flows:
            nodes.append(flow.target)
        return nodes

    def all_source_node(self):
        """
        返回当前集合中所有 flow 的 source
        :return:
        """
        nodes = []
        for flow in self.flows:
            nodes.append(flow.source)
        return nodes

    def __iter__(self):
        return iter(self.flows)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import models
from django.utils.translation import ugettext_lazy as _

from pipeline.conf import settings
from pipeline.contrib.external_plugins.models.base import FILE_SYSTEM, GIT, S3, ExternalPackageSource, package_source
from pipeline.contrib.external_plugins.utils.importer import FSModuleImporter, GitRepoModuleImporter, S3ModuleImporter


@package_source
class GitRepoSource(ExternalPackageSource):
    repo_raw_address = models.TextField(_("文件托管仓库链接"))
    branch = models.CharField(_("分支名"), max_length=128)

    @staticmethod
    def type():
        return GIT

    def importer(self):
        return GitRepoModuleImporter(
            name=self.name,
            repo_raw_url=self.repo_raw_address,
            branch=self.branch,
            modules=list(self.packages.keys()),
            proxy=settings.EXTERNAL_PLUGINS_SOURCE_PROXY,
            secure_only=settings.EXTERNAL_PLUGINS_SOURCE_SECURE_RESTRICT,
        )

    def details(self):
        return {"repo_raw_address": self.repo_raw_address, "branch": self.branch}


@package_source
class S3Source(ExternalPackageSource):
    service_address = models.TextField(_("对象存储服务地址"))
    bucket = models.TextField(_("bucket 名"))
    source_dir = models.TextField(_("源目录名"), default="")
    access_key = models.TextField(_("access key"))
    secret_key = models.TextField(_("secret key"))

    @staticmethod
    def type():
        return S3

    def importer(self):
        return S3ModuleImporter(
            name=self.name,
            modules=list(self.packages.keys()),
            service_address=self.service_address,
            bucket=self.bucket,
            source_dir=self.source_dir,
            access_key=self.access_key,
            secret_key=self.secret_key,
            secure_only=settings.EXTERNAL_PLUGINS_SOURCE_SECURE_RESTRICT,
        )

    def details(self):
        return {
            "service_address": self.service_address,
            "bucket": self.bucket,
            "source_dir": self.source_dir,
            "access_key": self.access_key,
            "secret_key": self.secret_key,
        }


@package_source
class FileSystemSource(ExternalPackageSource):
    path = models.TextField(_("文件系统路径"))

    @staticmethod
    def type():
        return FILE_SYSTEM

    def importer(self):
        return FSModuleImporter(name=self.name, modules=list(self.packages.keys()), path=self.path)

    def details(self):
        return {"path": self.path}



# Generated by Django 2.2.16 on 2021-06-24 15:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("pipeline", "0030_auto_20210607_1210"),
    ]

    operations = [
        migrations.AddField(
            model_name="pipelineinstance",
            name="is_expired",
            field=models.BooleanField(default=False, help_text="运行时被定期清理即为过期", verbose_name="是否已经过期"),
        ),
        migrations.AlterField(
            model_name="pipelineinstance",
            name="create_time",
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name="创建时间"),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from typing import Optional

from pipeline.eri.models import LogEntry


class HooksMixin:
    def pre_prepare_run_pipeline(
        self, pipeline: dict, root_pipeline_data: dict, root_pipeline_context: dict, subprocess_context: dict, **options
    ):
        """
        调用 pre_prepare_run_pipeline 前执行的钩子

        :param pipeline: 流程描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data: dict
        :param root_pipeline_context 根流程上下文
        :type root_pipeline_context: dict
        :param subprocess_context 子流程预置流程上下文
        :type subprocess_context: dict
        """

    def post_prepare_run_pipeline(
        self, pipeline: dict, root_pipeline_data: dict, root_pipeline_context: dict, subprocess_context: dict, **options
    ):
        """
        调用 pre_prepare_run_pipeline 后执行的钩子

        :param pipeline: 流程描述对象
        :type pipeline: dict
        :param root_pipeline_data 根流程数据
        :type root_pipeline_data: dict
        :param root_pipeline_context 根流程上下文
        :type root_pipeline_context: dict
        :param subprocess_context 子流程预置流程上下文
        :type subprocess_context: dict
        """

    def pre_pause_pipeline(self, pipeline_id: str):
        """
        暂停 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def post_pause_pipeline(self, pipeline_id: str):
        """
        暂停 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def pre_revoke_pipeline(self, pipeline_id: str):
        """
        撤销 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def post_revoke_pipeline(self, pipeline_id: str):
        """
        撤销 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def pre_resume_pipeline(self, pipeline_id: str):
        """
        继续 pipeline 前执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def post_resume_pipeline(self, pipeline_id: str):
        """
        继续 pipeline 后执行的钩子

        :param pipeline_id: 流程 ID
        :type pipeline_id: str
        """

    def pre_resume_node(self, node_id: str):
        """
        继续节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    def post_resume_node(self, node_id: str):
        """
        继续节点后执行的钩子

        :param node_id: [description]节点 ID
        :type node_id: str
        """

    def pre_pause_node(self, node_id: str):
        """
        暂停节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    def post_pause_node(self, node_id: str):
        """
        暂停节点后执行的钩子

        :param node_id: [description]节点 ID
        :type node_id: str
        """

    def pre_retry_node(self, node_id: str, data: Optional[dict]):
        """
        重试节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 重试时使用的节点执行输入
        :type data: Optional[dict]
        """

    def post_retry_node(self, node_id: str, data: Optional[dict]):
        """
        重试节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param data: 重试时使用的节点执行输入
        :type data: Optional[dict]
        """

    def pre_skip_node(self, node_id: str):
        """
        跳过节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    def post_skip_node(self, node_id: str):
        """
        跳过节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        """

    def pre_skip_exclusive_gateway(self, node_id: str, flow_id: str):
        """
        跳过分支网关前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_id: 跳过后选择的目标流 ID
        :type flow_id: str
        """

    def post_skip_exclusive_gateway(self, node_id: str, flow_id: str):
        """
        跳过分支网关后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_id: 跳过后选择的目标流 ID
        :type flow_id: str
        """

    def pre_skip_conditional_parallel_gateway(self, node_id: str, flow_ids: list, converge_gateway_id: str):
        """
        跳过条件并行网关前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_ids: 跳过后选择的目标流 ID 列表
        :type flow_ids: list
        :param converge_gateway_id: 目标汇聚网关 ID
        :type converge_gateway_id: str
        """

    def post_skip_conditional_parallel_gateway(self, node_id: str, flow_ids: list, converge_gateway_id: str):
        """
        跳过条件并行网关后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param flow_ids: 跳过后选择的目标流 ID 列表
        :type flow_ids: list
        :param converge_gateway_id: 目标汇聚网关 ID
        :type converge_gateway_id: str
        """

    def pre_forced_fail_activity(self, node_id: str, ex_data: str):
        """
        强制失败节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param ex_data: 写入节点执行数据的失败信息
        :type ex_data: str
        """

    def post_forced_fail_activity(self, node_id: str, ex_data: str, old_version: str, new_version: str):
        """
        强制失败节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param ex_data: 写入节点执行数据的失败信息
        :type ex_data: str
        :param old_version: 强制失败前的节点版本
        :type old_version: str
        :param new_version: 强制失败后的节点版本
        :type new_version: str
        """
        # 在强制失败刷新版本后更新已经记录的日志的版本
        LogEntry.objects.filter(node_id=node_id, version=old_version).update(version=new_version)

    def pre_callback(self, node_id: str, version: str, data: str):
        """
        回调节点前执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: str
        """

    def post_callback(self, node_id: str, version: str, data: str):
        """
        回调节点后执行的钩子

        :param node_id: 节点 ID
        :type node_id: str
        :param version: 节点执行版本
        :type version: str
        :param data: 回调数据
        :type data: str
        """

    def pre_retry_subprocess(self, node_id: str):
        """
        子流程重试前执行的钩子

        :param node_id: 子流程节点 ID
        :type node_id: str
        """

    def post_retry_subprocess(self, node_id: str):
        """
        子流程重试后执行的钩子

        :param node_id: 子流程节点 ID
        :type node_id: str
        """



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.core.constants import PE
from pipeline.utils.collections import FancyDict


class Data(object):
    def __init__(self, inputs=None, outputs=None):
        self.inputs = FancyDict(inputs or {})
        self.outputs = outputs or []

    def to_dict(self):
        base = {"inputs": {}, "outputs": self.outputs}

        for key, value in list(self.inputs.items()):
            base["inputs"][key] = value.to_dict() if isinstance(value, Var) else value

        return base


class Params(object):
    def __init__(self, params=None):
        self.params = FancyDict(params or {})

    def to_dict(self):
        base = {}

        for key, value in list(self.params.items()):
            base[key] = value.to_dict() if isinstance(value, Var) else value

        return base


class Var(object):
    PLAIN = PE.plain
    SPLICE = PE.splice
    LAZY = PE.lazy

    def __init__(self, type, value, custom_type=None):
        self.type = type
        self.value = value
        self.custom_type = custom_type

    def to_dict(self):
        base = {"type": self.type, "value": self.value}
        if self.type == self.LAZY:
            base["custom_type"] = self.custom_type

        return base


class DataInput(Var):
    def __init__(self, *args, **kwargs):
        super(DataInput, self).__init__(*args, **kwargs)

    def to_dict(self):
        base = super(DataInput, self).to_dict()
        base["is_param"] = True
        return base


class NodeOutput(Var):
    def __init__(self, source_act, source_key, *args, **kwargs):
        self.source_act = source_act
        self.source_key = source_key
        kwargs["value"] = None
        super(NodeOutput, self).__init__(*args, **kwargs)

    def to_dict(self):
        base = super(NodeOutput, self).to_dict()
        base["source_act"] = self.source_act
        base["source_key"] = self.source_key
        return base


class RewritableNodeOutput(Var):
    def __init__(self, source_act, *args, **kwargs):
        self.source_act = source_act
        kwargs["value"] = None
        super(RewritableNodeOutput, self).__init__(*args, **kwargs)

    def to_dict(self):
        base = super(RewritableNodeOutput, self).to_dict()
        base["source_act"] = self.source_act
        return base



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from .models import Signal

logger = logging.getLogger(__name__)


def set_valve_function(func):
    global __valve_function
    if __valve_function is not None:
        raise Exception("valve function can only be set once.")
    if not callable(func):
        raise Exception("valve function must be a callable object")

    __valve_function = func


def send(signal_mod, signal_name, **kwargs):
    if not __valve_function or not __valve_function():
        try:
            return getattr(signal_mod, signal_name).send(**kwargs)
        except Exception:
            raise
    else:
        Signal.objects.dump(signal_mod.__path__, signal_name, kwargs)
        return None


def open_valve(signal_mod):
    signal_list = Signal.objects.filter(module_path=signal_mod.__path__).order_by("id")
    response = []
    for signal in signal_list:
        try:
            response.append(getattr(signal_mod, signal.name).send(**signal.kwargs))
            signal.delete()
        except Exception:
            logger.error(
                "signal({} - {}) resend failed: {}".format(signal.module_path, signal.name, traceback.format_exc())
            )
    return response


def unload_valve_function():
    global __valve_function
    __valve_function = None


def valve_function():
    return __valve_function


__valve_function = None



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from pipeline.core.constants import PE
from pipeline.utils.uniqid import uniqid

__all__ = ["Element", "PE"]


class Element(object):
    def __init__(self, id=None, name=None, outgoing=None):
        self.id = id or uniqid()
        self.name = name
        self.outgoing = outgoing or []

    def extend(self, element):
        """
        build a connection from self to element and return element
        :param element: target
        :rtype: Element
        """
        self.outgoing.append(element)
        return element

    def connect(self, *args):
        """
        build connections from self to elements in args and return self
        :param args: target elements
        :rtype: Element
        """
        for e in args:
            self.outgoing.append(e)
        return self

    def converge(self, element):
        """
        converge all connection those diverge from self to element and return element
        :param element: target
        :rtype: Element
        """
        for e in self.outgoing:
            e.tail().connect(element)
        return element

    def to(self, element):
        return element

    def tail(self):
        """
        get tail element for self
        :rtype: Element
        """
        is_tail = len(self.outgoing) == 0
        e = self

        while not is_tail:
            e = e.outgoing[0]
            is_tail = len(e.outgoing) == 0

        return e

    def type(self):
        raise NotImplementedError()

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return "<{cls} {name}:{id}>".format(cls=type(self).__name__, name=self.name, id=self.id)



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

from django.db import transaction
from django_celery_beat.models import (
    IntervalSchedule,
    CrontabSchedule,
    PeriodicTask,
)

from pipeline.contrib.periodic_task.djcelery.models import (
    IntervalSchedule as DjCeleryIntervalSchedule,
    CrontabSchedule as DjCeleryCrontabSchedule,
    DjCeleryPeriodicTask,
)

BATCH_SIZE = 500


@transaction.atomic
def try_to_migrate_to_django_celery_beat():
    """
    try to migrate djcelery to django_celery_beat
    if django_celery_beat models has data, indicate that pipeline is first use in project
    (because old version pipeline is not compatible with django celery beat)
    so we will not do the migration works
    """
    if IntervalSchedule.objects.exists() or CrontabSchedule.objects.exists() or PeriodicTask.objects.exists():
        print("django_celery_beat in used, skip pipeline djcelery migration works")
        return

    # migrate IntervalScheudle
    old_intervals = DjCeleryIntervalSchedule.objects.all()
    new_intervals = []
    for oi in old_intervals:
        new_intervals.append(IntervalSchedule(id=oi.id, every=oi.every, period=oi.period))
    IntervalSchedule.objects.bulk_create(new_intervals, batch_size=BATCH_SIZE)
    print("[pipeline]migrate {} interval objects".format(len(new_intervals)))

    # migrate CrontabSchedule
    old_crontabs = DjCeleryCrontabSchedule.objects.all()
    new_crontabs = []
    for oc in old_crontabs:
        new_crontabs.append(
            CrontabSchedule(
                id=oc.id,
                minute=oc.minute,
                hour=oc.hour,
                day_of_week=oc.day_of_week,
                day_of_month=oc.day_of_month,
                month_of_year=oc.month_of_year,
                timezone=oc.timezone,
            )
        )
    CrontabSchedule.objects.bulk_create(new_crontabs, batch_size=BATCH_SIZE)
    print("[pipeline]migrate {} crontab objects".format(len(new_crontabs)))

    # migrate PeriodicTask
    old_tasks = DjCeleryPeriodicTask.objects.all()
    new_tasks = []
    for ot in old_tasks:
        new_tasks.append(
            PeriodicTask(
                id=ot.id,
                name=ot.name,
                task=ot.task,
                interval_id=ot.interval_id,
                crontab_id=ot.crontab_id,
                solar_id=None,
                clocked_id=None,
                args=ot.args,
                kwargs=ot.kwargs,
                queue=ot.queue,
                exchange=ot.exchange,
                routing_key=ot.routing_key,
                expires=ot.expires,
                enabled=ot.enabled,
                last_run_at=ot.last_run_at,
                total_run_count=ot.total_run_count,
                date_changed=ot.date_changed,
                description=ot.description,
            )
        )
    PeriodicTask.objects.bulk_create(new_tasks, batch_size=BATCH_SIZE)
    print("[pipeline]migrate {} periodic tasks".format(len(new_tasks)))



# -*- coding: utf-8 -*-
"""
Tencent is pleased to support the open source community by making 蓝鲸智云PaaS平台社区版 (BlueKing PaaS Community
Edition) available.
Copyright (C) 2017 THL A29 Limited, a Tencent company. All rights reserved.
Licensed under the MIT License (the "License"); you may not use this file except in compliance with the License.
You may obtain a copy of the License at
http://opensource.org/licenses/MIT
Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the License for the
specific language governing permissions and limitations under the License.
"""

import logging
import traceback

from pipeline.core.flow.gateway import ParallelGateway
from pipeline.engine.models import PipelineProcess, Status
from pipeline.exceptions import PipelineException

from .base import FlowElementHandler

logger = logging.getLogger("pipeline_engine")

__all__ = ["ParallelGatewayHandler"]


class ParallelGatewayHandler(FlowElementHandler):
    @staticmethod
    def element_cls():
        return ParallelGateway

    def handle(self, process, element, status):
        targets = element.outgoing.all_target_node()
        children = []

        for target in targets:
            try:
                child = PipelineProcess.objects.fork_child(
                    parent=process, current_node_id=target.id, destination_id=element.converge_gateway_id
                )
            except PipelineException as e:
                logger.error(traceback.format_exc())
                Status.objects.fail(element, str(e))
                return self.HandleResult(next_node=None, should_return=True, should_sleep=True)

            children.append(child)

        process.join(children)

        Status.objects.finish(element)

        return self.HandleResult(next_node=None, should_return=True, should_sleep=True)
